# SPDX-License-Identifier: MIT
"""Move files to directories based on matching attributes."""

import argparse
import builtins
import logging
import os
import re
import sys

from collections.abc import (
    Generator,
    Iterable,
    Mapping,
    MutableMapping,
    MutableSequence,
    Set,
)
from pathlib import Path
from typing import Any, Generic, Self, TypeVar

from fnattr.util import log, nested
from fnattr.util.config import read_cmd_configs_and_merge_options
from fnattr.util.typecheck import needtype
from fnattr.vljum.m import M
from fnattr.vljumap import enc

T = TypeVar('T')
V = TypeVar('V')

class Lazy(Generic[T]):
    """Holds a value."""

    def __init__(self, v: T) -> None:
        self.value: T = v

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.value!r})'

def lazify(a: Any) -> dict | list | set | Lazy:  # noqa: any-type
    if isinstance(a, Mapping):
        return {k: lazify(v) for k, v in a.items()}
    if isinstance(a, Set):
        # Pyright doesn't understand that lazify(v) is hashable if v is.
        return {lazify(v) for v in a}
    if isinstance(a, Iterable) and not isinstance(a, str):
        return [lazify(v) for v in a]
    return Lazy(a)

def zalify(
        a: Lazy[T] | T,
        values: Mapping) -> MutableMapping | MutableSequence | list | set | T:
    """Evaluate Lazy values in place."""
    if isinstance(a, Lazy):
        return zalify(a.value, values)
    if isinstance(a, MutableMapping):
        for k, v in a.items():
            a[k] = zalify(v, values)
        return a
    if isinstance(a, Set):
        return {zalify(v, values) for v in a}
    if isinstance(a, str):
        return zalify_str(a, values)
    if isinstance(a, MutableSequence):
        for i, v in enumerate(a):
            a[i] = zalify(v, values)
        return a
    if isinstance(a, Iterable):
        return [zalify(v, values) for v in a]
    return a

def zalify_str(s: str, values: Mapping) -> Any:
    # print(f'zalify_str {s=}')
    xl, xr = ('«', '»')
    if xl not in s:
        return s
    if s[0] == xl and s[-1] == xr:
        return zalify(nested.dget(values, s[1 :-1], None), values)
    r = []
    while xl in s:
        i = s.index(xl)
        j = s.index(xr, i)
        k = s[i + 1 : j]
        # print(f'     inner {k=}')
        v = nested.dget(values, k, '')
        # print(f'         ↦ {v=}')
        v = zalify(v, values)
        # print(f'         ↦ {v=}')
        r.append(s[: i])
        r.append(str(v))
        s = s[j + 1 :]
    r.append(s)
    return ''.join(r)

def flatten_list(a: Any) -> Generator:
    if isinstance(a, list | set | frozenset):
        for i in a:
            yield from flatten_list(i)
        return
    yield a

def zalify_to_frozenset(v: Any, values: Mapping) -> frozenset:
    t = list(flatten_list(zalify(v, values)))
    return frozenset(list(t))

FrozenSets = dict[str, frozenset]

class Destinations:
    """Record of target directories and associated conditions."""

    builtins = {
        k: getattr(builtins, k)
        for k in (
            'False',
            'None',
            'True',
            'all',
            'any',
            'frozenset',
            'len',
        )
    } | {
        'set': frozenset,
    }
    sets_re = re.compile('‹([^"›]+)›')

    def __init__(self, dests: list, defs: dict, sets: dict) -> None:
        self.values = {
            'def': defs,
            'set': sets,
            'env': os.environ,
        }
        self.singleset: dict[str, dict[frozenset, str]] = {}
        self.conditionals: list[dict] = []
        for k, v in sets.items():
            sets[k] = zalify_to_frozenset(v, self.values)
        for dest in dests:
            directory = zalify(dest['directory'], self.values)
            dest['directory'] = directory
            if 'set' in dest:
                sets = {
                    k: zalify_to_frozenset(v, self.values)
                    for k, v in dest['set'].items()
                }
                if len(sets) == 1 and 'condition' not in dest:
                    # Pure single set match.
                    k, v = list(sets.items())[0]
                    nested.nset(self.singleset, [k, frozenset(v)], directory)
                    continue
                dest['set'] = sets
            self.conditionals.append(dest)

    @classmethod
    def from_config(cls, config: Mapping) -> Self:
        defs: dict = needtype(lazify(config.get('def', {})), dict)
        sets: dict = needtype(lazify(config.get('set', {})), dict)
        dests: list = needtype(lazify(config.get('destination', [])), list)
        return cls(dests, defs, sets)

    def match(self, m: M) -> Path | None:
        msets: FrozenSets = {}
        if dst := self._match_single_set(m, msets):
            return dst
        if dst := self._match_conditionals(m, msets):
            return dst
        return None

    def _match_single_set(self, m: M, msets: FrozenSets) -> Path | None:
        for k, s in self.singleset.items():
            mset = mget(m, msets, k)
            if mset in s:
                return Path(s[mset])
        return None

    def _match_conditionals(self, m: M, msets: FrozenSets) -> Path | None:
        for c in self.conditionals:
            if dst := self._match_condition(m, msets, c):
                return dst
        return None

    def _match_condition(self, m: M, msets: FrozenSets, c: dict) -> Path | None:
        if 'set' in c:
            for k, s in c['set'].items():
                # Attribute must be a nonempty subset.
                mset = mget(m, msets, k)
                if not (mset and mset.issubset(s)):
                    return None
        condition = c['condition']
        if isinstance(condition, Lazy):
            condition = zalify(condition, self.values)
            if isinstance(condition, str):
                condition = self.sets_re.sub('(sets["\\1"])', condition)
            c['condition'] = condition
        if isinstance(condition, str):
            for k in m:
                mget(m, msets, k)
            logging.debug('condition: %s', condition)
            r = eval(condition, {'sets': msets, '__builtins__': self.builtins})
        else:
            r = condition
        if bool(r):
            return Path(c['directory'])
        return None

def mget(m: M, msets: FrozenSets, key: str) -> frozenset:
    if key not in msets:
        msets[key] = frozenset(str(a) for a in m.get(key, []))
    return msets[key]

def rename(m: M, *, dryrun: bool = False) -> bool:
    try:
        m.rename(dryrun=dryrun, dedup=True)
    except FileExistsError:
        logging.error('file exists: %s', m.filename())
        return False
    return True

def run(argv: list[str] | None = None) -> int:
    if argv is None:
        argv = sys.argv
    cmd = Path(argv[0]).stem
    parser = argparse.ArgumentParser(
        prog=cmd, description='Move files according to file name attributes')
    parser.add_argument(
        '--config',
        '-c',
        metavar='FILE',
        type=str,
        action='append',
        help='Configuration file.')
    parser.add_argument(
        '--decoder',
        '-d',
        metavar='DECODER',
        type=str,
        choices=enc.decoder.keys(),
        help='Default string decoder.')
    parser.add_argument(
        '--dryrun',
        '-n',
        default=False,
        action='store_true',
        help='Do not actually rename.')
    parser.add_argument(
        '--map',
        '-m',
        metavar='FILE',
        type=str,
        action='append',
        help='Renaming map file.')
    parser.add_argument(
        '--log-level',
        '-L',
        metavar='LEVEL',
        type=str,
        choices=log.CHOICES,
        default='WARNING')
    parser.add_argument(
        'file',
        metavar='FILENAME',
        type=str,
        nargs=argparse.REMAINDER,
        default=[],
        help='File name(s).')
    args = parser.parse_args(argv[1 :])
    log.level(cmd, args.log_level, dryrun=args.dryrun)
    config, options = read_cmd_configs_and_merge_options(
        cmd, args.config, args, decoder='v3')
    M.configure_options(options)
    #   M.configure_sites(config.get('site', {}))

    d = Destinations.from_config(config)
    for file in args.file:
        m = M().file(file)
        if not (dst := d.match(m)):
            logging.info('no match: %s', file)
            continue
        rename(m.with_dir(dst), dryrun=args.dryrun)

    return 0

def main(argv: list[str] | None = None) -> int:
    try:
        return run(argv)
    except Exception as e:
        logging.error('Unhandled exception: %s%s', type(e).__name__, e.args)
        if logging.getLogger().getEffectiveLevel() < logging.INFO:
            raise
        return 1

if __name__ == '__main__':
    sys.exit(main())



# SPDX-License-Identifier: MIT

import argparse
import hashlib
import logging
import os
import re
import sys

from pathlib import Path
from typing import Any

from fnattr.util.config import read_cmd_configs_and_merge_options
from fnattr.util.sqlite import SQLite
from fnattr.vlju.types.all import DOI
from fnattr.vljum.m import M
from fnattr.vljumap import enc

COLUMNS = [
    'Title',
    'Subtitle',
    'Author',
    'Editor',
    'Edition',
    'Copyright Year',
    'Publication Year',
    'ISBN#',
    'LCCN#',
    'ISSN#',
    'DOI',
    'Series Number',
    'LoC Classification',
]

def destination(p: Path, row, options: dict[str, Any]) -> Path | None:
    logging.debug('row=%s', repr(row))
    m = M()

    for k in ('Title', 'Subtitle'):
        if t := row.get(k):
            m.add('title', fix_title(t, options))

    for k, f in (('Author', 'a'), ('Editor', 'ed')):
        if t := row.get(k):
            for a in t.split(';'):
                m.add(f, a.strip().strip(','))

    for k, f in (('ISBN#', 'isbn'), ('LCCN#', 'lccn'), ('DOI', 'doi'),
                 ('ISSN#', 'issn'), ('Edition', 'edition'),
                 ('LoC Classification', 'loc'), ('Series Number', 'v')):
        if t := row.get(k):
            m.add(f, t)

    for k in ('Publication Date', 'Copyright Year'):
        if t := row.get(k):
            m.add('date', fix_date(t))

    subdir = destination_dir(m)
    dstdir = options['destination'] / subdir
    maxlen = min(options['PATH_MAX'] - len(str(dstdir)) - 1,
                 options['FILE_MAX']) - len(p.suffix)
    logging.debug("DIR:  '%s'", dstdir)
    file = destination_file(m, maxlen)
    if file is None:
        return None
    file += p.suffix
    logging.debug("FILE: '%s'", file)
    dst = Path(dstdir) / Path(file)
    logging.debug("DST:  '%s'", dst)
    return dst

def fix_title(s: str, options: dict[str, Any]) -> str:
    if options['the'] == 'start' and s.endswith(', The'):
        s = 'The ' + s[:-5]
    elif options['the'] == 'end' and s.startswith('The '):
        s = s[4 :] + ', The'
    return smarten_up(s)

def fix_date(s: str) -> str:
    y, n = re.subn(r'.*\b([12]\d{3})\b.*', r'\1', s)
    if n:
        s = y
    return s

def smarten_up(s: str) -> str:
    if '--' in s:
        s = s.replace('--', ' — ')
    return ' '.join(s.split())

def destination_dir(m: M) -> str:
    if 'loc' in m:
        t = str(m['loc'][0])
        b, n = re.subn(r'CPB Box no. (\d+).*', r'\1', t)
        if n:
            b = f'{int(b):04}'
            return f'lc/_CPB_Box_/{b[:2]}xx'
        r = ''
        while t and t[0].isalpha():
            r += t[0]
            t = t[1 :]
        r = r[: 2].upper()
        if r:
            return f'lc/{r}'

    if 'isbn' in m:
        t = str(m['isbn'][0])
        return f'isbn/{(int(t[:6]) - 978000):03}'

    if 'doi' in m:
        t = m['doi'][0]
        assert isinstance(t, DOI)
        return f'doi/{t.prefix()}'

    return 'other'

def destination_file(m: M, maxlen: int) -> str | None:
    keys = [
        k for k in ('title', 'a', 'ed', 'v', 'isbn', 'doi', 'lccn', 'issn',
                    'edition', 'date') if k in m
    ]
    m = m.submap(keys)

    encoder = m.encoder.get()
    s = encoder.encode(m)
    if len(s) <= maxlen:
        return s
    logging.debug('long: %d>%d %s', len(s), maxlen, s)

    for k in ('date', 'edition', 'ed'):
        if k in keys:
            m.remove(k)
            s = encoder.encode(m)
            if len(s) <= maxlen:
                return s
            logging.debug('long: %d>%d %s', len(s), maxlen, s)

    for k in ('title', 'a'):
        while len(m[k]) > 1:
            m[k].pop()
            s = encoder.encode(m)
            if len(s) <= maxlen:
                return s
            logging.debug('long: %d>%d %s', len(s), maxlen, s)

    return None

def sha1(p: Path) -> str | None:
    if p.is_file():
        try:
            with p.open('rb') as f:
                return hashlib.file_digest(f, hashlib.sha1).hexdigest()
        except OSError as e:
            logging.error(e)
            return None

    if p.is_dir():
        files = []
        for dirpath, _, filenames in os.walk(p):
            for file in filenames:
                if file == '.DS_Store':
                    continue
                files.append(f'{dirpath}/{file}')
        files.sort()
        h = hashlib.sha1()
        for file in files:
            with Path(file).open('rb') as f:
                h = hashlib.file_digest(f, lambda: h)
        return h.hexdigest()

    logging.error('%s not found', p)
    return None

def dict_factory(cursor, row) -> dict:
    return dict(zip((column[0] for column in cursor.description), row))

def main(argv):
    cmd = Path(argv[0]).stem
    r = 0

    parser = argparse.ArgumentParser(prog=cmd, description='TODO')
    parser.add_argument(
        '--config',
        '-c',
        metavar='FILE',
        type=str,
        action='append',
        help='Configuration file.')
    parser.add_argument(
        '--db',
        '--database',
        '-d',
        metavar='DB',
        type=str,
        help='SQLite database file.')
    parser.add_argument(
        '--table',
        '-t',
        metavar='TABLE',
        type=str,
        help='SQLite database table.')
    parser.add_argument(
        '--dedup',
        action='store_true',
        help='Remove if destination is identical.')
    parser.add_argument(
        '--destination',
        '--root',
        '-r',
        metavar='DIR',
        type=str,
        help='Destination root.')
    parser.add_argument('--dryrun', '-n', action='store_true')
    parser.add_argument(
        '--encoder',
        '-e',
        metavar='ENCODER',
        type=str,
        choices=enc.encoder.keys(),
        help='Filename encoder.')
    parser.add_argument(
        '--ignore-hash-collision',
        '-I',
        action='store_true',
        help='Ignore hash collisions; use first entry.')
    parser.add_argument(
        '--log-level',
        '-L',
        metavar='LEVEL',
        type=str,
        choices=[c for c in logging.getLevelNamesMapping() if c != 'NOTSET'],
        default='INFO')
    parser.add_argument('--sha', action='store_true', help='Print SHA1 only.')
    parser.add_argument(
        '--the',
        choices=('start', 'end', 'none'),
        help='Move ‘The’ in titles.',
    )
    parser.add_argument(
        'files',
        metavar='FILE',
        type=str,
        nargs='+',
        help='Files to rename',
    )
    args = parser.parse_args(argv[1 :])

    logging.basicConfig(
        level=getattr(logging, args.log_level.upper()),
        format=f'{cmd}: %(levelname)s: %(message)s')

    _config, options = read_cmd_configs_and_merge_options(
        cmd,
        args.config,
        args,
        db=None,
        table='ebooks',
        destination='.',
        the='start',
    )

    if not args.sha and not args.db:
        logging.error('--db is required')
        return 1

    options['destination'] = Path(options['destination'])
    options['PATH_MAX'] = os.pathconf(options['destination'], 'PC_PATH_MAX')
    options['FILE_MAX'] = os.pathconf(options['destination'], 'PC_NAME_MAX')

    M.configure_options(options)

    if args.db:
        db = SQLite(args.db).connect()
        db.connection().row_factory = dict_factory
    else:
        db = None

    for file in args.files:

        logging.debug('From: %s', file)
        src = Path(file)

        sha = sha1(src)
        if sha is None:
            r = 1
            continue
        if args.sha:
            print(f'{sha} {file}')
            continue
        logging.debug('sha=%s', sha)
        if not db:
            continue

        cursor = db.load(options['table'], *COLUMNS, SHA1=sha)
        rows = cursor.fetchall()
        if len(rows) == 0:
            logging.error("%s: No entry for '%s'", cmd, src)
            r = 1
            continue
        if len(rows) > 1:
            if args.ignore_hash_collision:
                logging.warning("Hash collision for '%s': %s", src, sha)
            else:
                logging.error("Hash collision for '%s': %s", src, sha)
                logging.debug('%s', rows[0])
                logging.debug('%s', rows[1])
                r = 1
                continue

        dst = destination(src, rows[0], options)
        if dst is None:
            logging.error("%s: Failed to build a new name for '%s'", cmd, src)
            r = 1
            continue

        if dst.exists():
            if src.samefile(dst):
                continue
            if args.dedup:
                dstsha = sha1(dst)
                if dstsha == sha:
                    logging.info('%s: destination is identical: %s', cmd, dst)
                    if not args.dryrun and src.is_file():
                        src.unlink()
                    continue
                logging.debug('%s %s', sha, src)
                logging.debug('%s %s', dstsha, dst)
            logging.error('destination exists: %s', dst)
            r = 1
            continue

        if not args.dryrun:
            if not dst.parent.is_dir():
                dst.parent.mkdir(parents=True)
            src.rename(dst)

        logging.info('To:   %s', dst)

    if db:
        db.close()
    return r

if __name__ == '__main__':
    sys.exit(main(sys.argv))



# SPDX-License-Identifier: MIT
"""Nested dictionaries."""

from collections.abc import (
    Iterable,
    Mapping,
    MutableMapping,
    MutableSequence,
    MutableSet,
    Sequence,
    Set,
)
from typing import Any, TypeVar

T = TypeVar('T')
D = TypeVar('D')
MM = TypeVar('MM', bound=MutableMapping)

def nget(d: Mapping[T, Any], keys: Iterable[T], default) -> Any:
    """Get from nested dictionaries."""
    t: Any = d
    try:
        for k in keys:
            if k not in t:
                return default
            t = t[k]
    except (KeyError, TypeError):
        return default
    return t

def ngetor(d: Mapping[T, Any],
           keys: Iterable[T],
           default: D | None = None) -> Any | D | None:
    return nget(d, keys, default)

def nset(d: MutableMapping, keys: Iterable, value: Any) -> None:
    """Store in nested dictionaries."""
    ki = iter(keys)
    try:
        key: Any = next(ki)
    except StopIteration:
        raise KeyError(keys)
    while True:
        try:
            next_key = next(ki)
        except StopIteration:
            break
        if key not in d:
            d[key] = {}
        d = d[key]
        key = next_key
    d[key] = value

def nupdate(d: MM, s: Mapping) -> MM:
    """Update nested dictionaries."""
    for k, v in s.items():
        if k not in d:
            d[k] = v
            continue
        if isinstance(d[k], MutableMapping) and isinstance(v, Mapping):
            nupdate(d[k], v)
            continue
        if isinstance(d[k], MutableSequence) and isinstance(v, Sequence):
            d[k] += v
            continue
        if isinstance(d[k], MutableSet) and isinstance(v, Set):
            d[k] |= v
            continue
        if type(d[k]) is type(v):
            d[k] = v
            continue
        raise TypeError(k, d[k], v)
    return d

def dget(d: Mapping, keys: str, default: Any) -> Any:
    return nget(d, keys.split('.'), default)

def dgetor(d: Mapping, keys: str, default: Any | None = None) -> Any | None:
    return nget(d, keys.split('.'), default)

def dset(d: MutableMapping, keys: str, value: Any) -> None:
    nset(d, keys.split('.'), value)



# SPDX-License-Identifier: MIT
"""Configuration-file related utilities."""

import argparse
import contextlib
import logging
import os
import tomllib

from collections.abc import Iterable, Mapping
from pathlib import Path
from typing import Any, Self

from fnattr.util import nested

class Dirs(list[Path]):
    """Maintain a list of directories."""

    def add_env_dir(self, evar: str, default: Path | None = None) -> Self:
        """
        Add an environment-variable-based or default directory.

        Based on XDG conventions. If the environment variable named `evar`
        exists and contains a directory path, add it; otherwise, if the
        `default` exists, add that.
        """
        if evar in os.environ:
            p = Path(os.environ[evar])
            if p.is_dir():
                self.append(p)
        elif default and default.is_dir():
            self.append(default)
        return self

    def add_env_dirs(self, env: str, default: list[Path]) -> Self:
        """
        Add a list of environment-variable-based or default directories.

        Based on XDG conventions. If the environment variable named `evar`
        exists, treat it as a `:`-separated path and add any existing absolute
        directory. Otherwise, add any existing directory in `default`.
        """
        if env in os.environ:
            for d in os.environ[env].split(':'):
                p = Path(d)
                if p.is_dir() and p.is_absolute():
                    self.append(p)
        else:
            for p in default:
                if p.is_dir():
                    self.append(p)
        return self

    def find_first(self, p: Path | str) -> Path | None:
        """Return the first matching file in the directory list."""
        for i in self:
            d = i / p
            if d.exists():
                return d
        return None

def xdg_dirs(name: str,
             default_dir: str,
             default_paths: list[Path],
             home: Path | None = None) -> Dirs:
    """Obtain a list of XDG directories of the given kind."""
    if home is None:
        with contextlib.suppress(RuntimeError):
            home = Path.home()

    default_path = None if home is None else home / default_dir
    d = Dirs()
    d.add_env_dir(f'XDG_{name}_HOME', default_path)
    d.add_env_dirs(f'XDG_{name}_DIRS', default_paths)
    return d

def xdg_config_dirs() -> Dirs:
    return xdg_dirs('CONFIG', '.config', [Path('/etc/xdg')])

def xdg_config(filename: Path | str) -> Path | None:
    return xdg_config_dirs().find_first(filename)

def read_toml_config(file: Path | str) -> dict | None:
    with Path(file).open('rb') as f:
        try:
            return tomllib.load(f)
        except tomllib.TOMLDecodeError as e:
            logging.error('%s: %s', file, str(e))
            return None

def read_configs(args: Iterable[Path | str]) -> dict:
    config: dict[str, Any] = {}
    for p in args:
        if c := read_toml_config(p):
            nested.nupdate(config, c)
    return config

def read_xdg_configs(files: Iterable[Path]) -> dict:
    config: dict[str, Any] = {}
    for file in files:
        if (cf := xdg_config(file)) and (c := read_toml_config(cf)):
            nested.nupdate(config, c)
    return config

def read_cmd_configs(cmds: str | Iterable[str],
                     config_files: Iterable[Path | str]) -> dict:
    paths = []
    names = ['vlju', cmds] if isinstance(cmds, str) else ['vlju', *cmds]
    for i in names:
        paths.append(Path(f'{i}.toml'))
        paths.append(Path(f'fnattr/{i}.toml'))
    config = read_xdg_configs(paths)
    if config_files:
        nested.nupdate(config, read_configs(config_files))
    return config

def merge_options(options: dict[str, Any] | None, args: argparse.Namespace,
                  **kwargs) -> dict[str, Any]:
    if options is None:
        options = {}
    for k, d in kwargs.items():
        option = d.get('option', k) if isinstance(d, Mapping) else k
        default = d.get('default') if isinstance(d, Mapping) else d
        if (a := getattr(args, k)) is not None:
            nested.dset(options, option, a)
        elif k not in options and default is not None:
            nested.dset(options, option, default)
    return options

def read_cmd_configs_and_merge_options(cmds: str | Iterable[str],
                                       config_files: Iterable[Path | str],
                                       args: argparse.Namespace,
                                       **kwargs) -> tuple[dict, dict]:
    config = read_cmd_configs(cmds, config_files)
    options = merge_options(config.get('option'), args, **kwargs)
    return config, options



# SPDX-License-Identifier: MIT
"""Rename files using Danbooru metadata."""

import argparse
import hashlib
import logging
import sys

from collections.abc import Callable, Generator, Mapping
from pathlib import Path

import pybooru  # type: ignore[import]

from fnattr.extra.fnaffle import Destinations, rename
from fnattr.util import log, nested
from fnattr.util.config import read_cmd_configs_and_merge_options
from fnattr.vljum.m import M
from fnattr.vljumap import enc

def split_tags(tags: str) -> Generator[str, None, None]:
    return (s.replace('_', ' ') for s in tags.split())

def add_tags(m: M,
             key: str,
             tags: str,
             cmap: Mapping,
             cf: Callable[[str], str] | None = None) -> None:
    if not tags:
        return
    for tag in split_tags(tags):
        if cmap and tag in cmap:
            v = cmap[tag]
        elif cf:
            v = cf(tag)
        else:
            v = tag
        m.add(key, v)

def capitalize_tag(s: str) -> str:
    if (n := s.find('(')) > 0:
        s = s[: n].strip()
    return ' '.join(t.capitalize() for t in s.split())

def file_md5(file: str | Path) -> str | None:
    try:
        with Path(file).open('rb') as f:
            return hashlib.file_digest(f, hashlib.md5).hexdigest()
    except OSError as e:
        logging.error(e)
        return None

def dan_fetch_md5(d: pybooru.Danbooru, md5: str) -> dict | None:
    try:
        return d.post_list(md5=md5)
    except pybooru.exceptions.PybooruHTTPError:
        return None

def danboorize(m: M, p: dict, defs: Mapping) -> None:
    if 'id' in p:
        m.reset('dan', str(p['id']))
    add_tags(m, 'c', str(p.get('tag_string_character', '')),
             nested.nget(defs, ['tag', 'c'], []), capitalize_tag)
    add_tags(m, 'a', str(p.get('tag_string_artist', '')),
             nested.nget(defs, ['tag', 'a'], []), capitalize_tag)
    if (tags := defs.get('tag')):
        for tag in split_tags(p.get('tag_string', '')):
            if tag in tags and isinstance((kv := tags[tag]), str):
                kv = tags[tag]
                if '=' in kv:
                    k, v = kv.split('=', 1)
                else:
                    k = kv
                    v = ''
                m.add(k, v)
    if (source := p.get('source')):
        k, v = m.from_site_url(source)
        if k:
            m.add(k, v)
    if (pixiv := p.get('pixiv_id')) and 'pixiv' not in m:
        m.add('pixiv', str(pixiv))
    m.order()

def run(argv: list[str] | None = None) -> int:
    if argv is None:
        argv = sys.argv
    cmd = Path(argv[0]).stem
    parser = argparse.ArgumentParser(
        prog=cmd, description='Rename files according to Danbooru metadata')
    parser.add_argument(
        '--config',
        '-c',
        metavar='FILE',
        type=str,
        action='append',
        help='Configuration file.')
    parser.add_argument(
        '--decoder',
        '-d',
        metavar='DECODER',
        type=str,
        choices=enc.decoder.keys(),
        help='File name decoder.')
    parser.add_argument(
        '--encoder',
        '-e',
        metavar='DECODER',
        type=str,
        choices=enc.encoder.keys(),
        help='File name encoder.')
    parser.add_argument(
        '--dryrun',
        '-n',
        default=False,
        action='store_true',
        help='Do not actually rename.')
    parser.add_argument(
        '--md5',
        default=False,
        action='store_true',
        help='Arguments are MD5 hashes rather than file names.')
    parser.add_argument(
        '--merge',
        '-m',
        default=True,
        action=argparse.BooleanOptionalAction,
        help='Merge with existing attributes.')
    parser.add_argument(
        '--log-level',
        '-L',
        metavar='LEVEL',
        type=str,
        choices=log.CHOICES,
        default='info')
    parser.add_argument(
        '--url',
        '-u',
        metavar='URL',
        type=str,
        default='https://danbooru.donmai.us/')
    parser.add_argument('--user', '-U', metavar='USER', type=str)
    parser.add_argument(
        '--token', '-T', metavar='TOKEN', type=str, help='API token.')

    parser.add_argument(
        '-Q', default=False, action='store_true', help='Quit early.')

    parser.add_argument(
        'file',
        metavar='FILENAME',
        type=str,
        nargs=argparse.REMAINDER,
        default=[],
        help='File name(s).')
    args = parser.parse_args(argv[1 :])

    log.level(cmd, args.log_level, dryrun=args.dryrun)
    config, options = read_cmd_configs_and_merge_options(
        ['fnaffle', cmd],
        args.config,
        args,
        decoder='v3',
        encoder='v3',
        user={'option': 'provider.danbooru.user'},
        token={'option': 'provider.danbooru.token'},
    )
    M.configure_options(options)
    #   M.configure_sites(config.get('site', {}))

    tags: dict[str, str] = {}
    for k in (['global'], ['provider', 'danbooru']):
        if (m := nested.ngetor(config, [*k, 'tag'])):
            tags = nested.nupdate(tags, m)

    d = Destinations.from_config(config)
    d.values['tag'] = tags

    s = pybooru.Danbooru(
        site_url=args.url,
        username=nested.nget(options, ['provider', 'danbooru', 'user'], ''),
        api_key=nested.nget(options, ['provider', 'danbooru', 'token'], ''))
    for file in args.file:
        m = M()
        if args.merge:
            m.file(file)
        else:
            m.set_path(file)
        if args.md5:
            md5 = file
        elif not (md5 := file_md5(file)):
            continue
        logging.debug('MD5 %s %s', md5, file)

        if not (p := dan_fetch_md5(s, md5)):
            logging.error('No match for: %s', file)
            continue

        danboorize(m, p, d.values)
        if dst := d.match(m):
            m.with_dir(dst)
        rename(m, dryrun=args.dryrun or args.md5)

    return 0

def main(argv: list[str] | None = None) -> int:
    try:
        return run(argv)
    except Exception as e:
        logging.error('Unhandled exception: %s%s', type(e).__name__, e.args)
        if logging.getLogger().getEffectiveLevel() < logging.INFO:
            raise
        return 1

if __name__ == '__main__':
    sys.exit(main())



# SPDX-License-Identifier: MIT
"""Logging utilities."""

import logging

CHOICES = [c.lower() for c in logging.getLevelNamesMapping() if c != 'NOTSET']

def level(cmd: str,
          s: str | int,
          *,
          dryrun: bool = False,
          dryrun_level: int = logging.INFO) -> int:
    log_level = getattr(logging, s.upper()) if isinstance(s, str) else s
    if dryrun and log_level > dryrun_level:
        log_level = dryrun_level
    logging.basicConfig(
        level=log_level, format=f'{cmd}: %(levelname)s: %(message)s')
    return log_level
